"""
Devanagari plugin for manim

Attribute
---------

* _FONT_NAME
* devanagari
* __version__

Classes
-------

* `Deva_Tex(tex_string, ....)` is normal mode (`Tex`)
* `Deva_MathTex(tex_string, ....)` is math mode (`MathTex`)
* `Deva_Text(text_string, ...)` is non-LaTeX string (`Text`)
* `Deva_MarkupText(text_string, ...)` is non-LaTeX string (`MarkupText`)
* `Question_Header(question_no,...)` is Question Header
* `Solution_Header(ans, ...)` is Solution
* `Cancel(...)` is cancel mobject

"""

import manimpango
from manim import *

__version__ = "2.0.0"


_Devanagari = TexTemplate(
    tex_compiler="xelatex",
    output_format=".xdv",
    documentclass="\\documentclass[preview]{standalone}",
    preamble="\\usepackage{fontspec}\n\\usepackage{polyglossia}\n\\usepackage{cancel}\n\\setmainlanguage{english}\n\\setotherlanguage{hindi}\\setmainfont[Script=Devanagari]{Noto Sans}\n\\usepackage{amsmath}\n\\usepackage{amssymb}",
)


_FONT_NAME = "sans-serif"
# _FONT_NAME="Noto Sans"

_SET_COLOR = BLACK
_SET_QUESTION_COLOR = RED
_SET_SOLUTION_COLOR = GREEN_E

_SET_FONT_SIZE = 35


_LIGHT_MODE = False
if _LIGHT_MODE:
    config.background_color = WHITE
    _SET_COLOR = WHITE

_SET_TEX_FONT_SIZE = _SET_FONT_SIZE - 5
_SET_TEXT_FONT_SIZE = _SET_TEX_FONT_SIZE - 4
_SET_MARKUPTEXT_FONT_SIZE = _SET_TEX_FONT_SIZE - 5


class Tex(Tex):
    def __init__(self, *args, font_size=_SET_TEX_FONT_SIZE, color=_SET_COLOR, **kwargs):
        super().__init__(*args, color=color, font_size=font_size, **kwargs)


class Deva_Tex(Tex):
    """Devanagari and English string compiled with LaTeX in normal mode."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, tex_template=_Devanagari, **kwargs)


class MathTex(MathTex):
    def __init__(self, *args, font_size=_SET_TEX_FONT_SIZE, color=_SET_COLOR, **kwargs):
        super().__init__(*args, color=color, font_size=font_size, **kwargs)


class Deva_MathTex(MathTex):
    """Devanagari and English string compiled with LaTeX in math mode."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, tex_template=_Devanagari, **kwargs)


class Text(Text):
    """Display (non-LaTeX) text rendered using Pango ."""

    def __init__(
        self, *args, font_size=_SET_TEXT_FONT_SIZE, color=_SET_COLOR, **kwargs
    ):
        super().__init__(*args, color=color, font_size=font_size, **kwargs)


class Deva_Text(Text):
    """Display (non-LaTeX) text rendered using Pango ."""

    def __init__(self, *args, font=_FONT_NAME, **kwargs):
        super().__init__(*args, font=font, **kwargs)


class Deva_MarkupText(MarkupText):
    """
    Display (non-LaTeX) text rendered using Pango .

    Text objects behave like a VGroup-like iterable of all characters in the given text. In particular, slicing is possible.

    What is PangoMarkup?

    PangoMarkup is a small markup language like html and it helps you avoid using "range of characters" while coloring or styling a piece a Text. You can use this language with ~.MarkupText.

    A simple example of a marked-up string might be

    <span foreground="blue" size="x-large">Blue text</span> is <i>cool</i>!"
    and it can be used with ~.MarkupText as

    .. manim:: MarkupExample :save_last_frame:
    """

    def __init__(
        self, *args, color=_SET_COLOR, font_size=_SET_MARKUPTEXT_FONT_SIZE, **kwargs
    ):
        super().__init__(*args, color=color, font_size=font_size, **kwargs)


class Deva_MarkupText(MarkupText):
    """
    Display (non-LaTeX) text rendered using Pango .

    Text objects behave like a VGroup-like iterable of all characters in the given text. In particular, slicing is possible.

    What is PangoMarkup?

    PangoMarkup is a small markup language like html and it helps you avoid using "range of characters" while coloring or styling a piece a Text. You can use this language with ~.MarkupText.

    A simple example of a marked-up string might be

    <span foreground="blue" size="x-large">Blue text</span> is <i>cool</i>!"
    and it can be used with ~.MarkupText as

    .. manim:: MarkupExample :save_last_frame:
    """

    def __init__(self, *args, font=_FONT_NAME, **kwargs):
        super().__init__(*args, font=font, **kwargs)


class Question_Header(Text):
    """Question Header
    Args:
        question_no (int): Question Number
    """

    def __init__(
        self,
        question_no,
        font=_FONT_NAME,
        font_size=30,
        color=_SET_QUESTION_COLOR,
        weight=BOLD,
        **kwargs,
    ):

        super().__init__(
            "प्रश्‍न {}:".format(question_no),
            font=font,
            font_size=font_size,
            color=color,
            weight=weight,
            **kwargs,
        )


class Solution(Text):
    """Question Header
    Args:
        ans (bool): Default True is 'उत्तर' and False is 'हल'

    """

    def __init__(
        self,
        font=_FONT_NAME,
        font_size=25,
        color=_SET_SOLUTION_COLOR,
        weight=BOLD,
        **kwargs,
    ):

        super().__init__(
            text="हल :",
            font=font,
            font_size=font_size,
            color=color,
            weight=weight,
            **kwargs,
        )


class Solution_Group(VGroup):
    r"""

    Solution_Group : Solution Group

    Structure
        [
            Group(Solution, line),
            *args
        ]

    """

    def __init__(
        self,
        line: str = Text(""),
        *args,
        solution_header_kwargs: dict = {},
        first_kwargs: dict = {},
        **kwargs,
    ):

        super().__init__(
            VGroup(Solution(**solution_header_kwargs), line, **first_kwargs).arrange(
                RIGHT
            ),
            *args,
            **kwargs,
        )


class Solution_VGroup(VGroup):
    r"""

    Solution_VGroup : Solution VGroup

    Structure
        [
            Solution,
            *args
        ]

    """

    def __init__(self, *args, solution_header_kwargs: dict = {}, **kwargs):
        super().__init__(Solution(**solution_header_kwargs), *args, **kwargs)


class Answer(Text):
    """Question Header
    Args:
        ans (bool): Default True is 'उत्तर' and False is 'हल'

    """

    def __init__(
        self,
        font=_FONT_NAME,
        font_size=25,
        color=_SET_SOLUTION_COLOR,
        weight=BOLD,
        **kwargs,
    ):

        super().__init__(
            text="उत्तर :",
            font=font,
            font_size=font_size,
            color=color,
            weight=weight,
            **kwargs,
        )


class Answer_Group(VGroup):
    r"""

    Solution_Group : Solution Group

    Structure
        [
            Group(Solution, line),
            *args
        ]

    """

    def __init__(
        self,
        line: str = Text(""),
        *args,
        solution_header_kwargs: dict = {},
        first_kwargs: dict = {},
        **kwargs,
    ):

        super().__init__(
            VGroup(Answer(**solution_header_kwargs), line, **first_kwargs).arrange(
                RIGHT
            ),
            *args,
            **kwargs,
        )


class Answer_VGroup(VGroup):
    r"""

    Solution_VGroup : Solution VGroup

    Structure
        [
            Solution,
            *args
        ]

    """

    def __init__(self, *args, solution_header_kwargs: dict = {}, **kwargs):
        super().__init__(Answer(**solution_header_kwargs), *args, **kwargs)


class Cancel(VGroup):
    r"""

    Cancel is `\cancel` same worked

    Args:
        mobject (Mobject): Manim Object

    """

    def __init__(
        self,
        mobject: Mobject | None = None,
        stroke_color: ParsableManimColor = RED,
        stroke_width: float = 2.0,
        scale_factor: float = 1.5,
        **kwargs,
    ) -> None:
        super().__init__(Line(UP + RIGHT, DOWN + LEFT), **kwargs)
        if mobject is not None:
            self.replace(mobject, stretch=True)
        self.scale(scale_factor)
        self.set_stroke(color=stroke_color, width=stroke_width)


class Question_Group(VGroup):
    r"""
    Question_Group : Question Group

    Structure
        [
            Question(id),
            *args
        ]

    """

    def __init__(self, id=0, *args, question_header_kwargs: dict = {}, **kwargs):
        super().__init__(
            Question_Header(question_no=id, **question_header_kwargs), *args, **kwargs
        )


def Str_Join(*args: str, space=False, **kwargs):
    args = [str(i) for i in args]
    return (" " if space else "\n").join(args)


Bookmark = lambda mark: "<bookmark mark='{}'/>".format(mark)



from manim import *
import manimpango
import functools


@functools.lru_cache(maxsize=None)
def font_list() -> list[str]:
    return manimpango.list_fonts()


def check_font(self, font: str, tuple_value=False) -> bool:
    fonts_list = font_list()
    font_variants = [font, font.lower(), font.capitalize(), font.title()]

    if font.lower() == "sans-serif":
        font = "sans"

    for variant in font_variants:
        if variant in fonts_list:
            return (variant, True) if tuple_value else True

    logger.warning(f"Font {font} not in {fonts_list}.")
    return (font, False) if tuple_value else False


def check_default_font(self, font: str) -> str:
    fonts = [
        font,
        "Lohit Devanagari",
        "Arial Unicode MS",
        "Noto Sans Devanagari",
        "Mangal",
        "Devanagari New",
        "Nirmala UI",
        "Hind",
    ]
    for f in fonts:
        f, check_font = check_font(f, tuple_value=True)
        if check_font:
            return f

    logger.error(f"Font {font} not in {fonts}.")



import textwrap
from manim import *
from typing import Iterable, Sequence
from typing_extensions import Self, TypeAlias
from manim.typing import Vector3D
from manim.mobject.opengl.opengl_vectorized_mobject import OpenGLVMobject


def Footer(
    width=None,
    stroke_opacity=0,
    fill_color=WHITE,
    fill_opacity=1,
    z_index=1,
    height=0.8,
) -> Rectangle:
    footer = Rectangle(
        stroke_opacity=stroke_opacity,
        fill_color=fill_color,
        fill_opacity=fill_opacity,
        z_index=z_index,
        width=width if width else config.frame_height,
        height=height,
    )
    footer.to_edge(DOWN, buff=0.0)
    return footer


class Text(Text):
    def __init__(
        self,
        text: str,
        align: str = "l",
        direction_align: str | None = None,
        *args,
        **kwargs,
    ):
        self.align = align
        self.direction_align = direction_align
        super().__init__(text, **kwargs)


class Paragraph(Paragraph):
    def __init__(
        self,
        *text: Sequence[str],
        wrap: bool = True,
        wrap_width: int = 50,
        align: str = "l",
        direction_align: str | None = None,
        **kwargs,
    ):
        self.wrap = wrap
        self.wrap_width = wrap_width
        self.direction_align = direction_align
        self.align = kwargs.get("alignment", align)
        self.original_text = map(self.wrap_text, text)
        super().__init__(
            *self.original_text,
            **kwargs,
        )

    def wrap_text(self, text: str) -> str:
        wrapped_text = textwrap.fill(text, width=self.wrap_width)

        # Justify lines based on alignment using match-case
        match self.align:
            case "r" | "right":
                justify_func = str.rjust
            case "c" | "center":
                justify_func = str.center
            case "l" | "left":  # Support for both 'l' and 'left'
                justify_func = str.ljust
            case _:
                justify_func = str.ljust  # Default to left justify
        # Default to left justify
        return "\n".join(
            justify_func(line, self.wrap_width) for line in wrapped_text.splitlines()
        )


class Tex(Tex):
    def __init__(self, *tex_strings, align="l", **kwargs):
        self.align = align
        super().__init__(*tex_strings, **kwargs)


class MathTex(MathTex):
    def __init__(self, *tex_strings, align="l", **kwargs):
        self.align = align
        super().__init__(*tex_strings, **kwargs)


class MarkupText(MarkupText):
    def __init__(self, text: str, align="l", **kwargs):
        self.align = align
        super().__init__(text, **kwargs)


class Notebook(VMobject):
    def __init__(self, *vmobjects, **kwargs):
        super().__init__(**kwargs)
        self.add(*vmobjects)
        self.notebook_arrange(aligned_edge=LEFT)
        self.to_corner(UL, buff=0.25)

    def notebook_arrange(
        self,
        direction: Vector3D = DOWN,
        buff: float = DEFAULT_MOBJECT_TO_MOBJECT_BUFFER,
        center: bool = True,
        **kwargs,
    ) -> Self:
        for m1, m2 in zip(self.submobjects, self.submobjects[1:]):
            m2.next_to(m1, direction, buff, **kwargs)

            align = (m1.align if hasattr(m1, "align") else m1.name).lower()
            match align:
                case "cue_column" | "cue-column" | "cue":
                    m1.to_edge(LEFT, buff=5 + buff)
                case "r" | "right":
                    m1.to_edge(RIGHT, buff=-(5 - buff))
                case "c" | "center":
                    m1.set_x(buff + 5)

        if center:
            self.center()
        return self


class QGroup(VMobject):
    def __init__(
        self,
        *vmobjects,
        question_no=0,
        lang="en",
        style={"font_size": 25, "color": GREEN_E, "weight": BOLD},
        **kwargs,
    ):
        super().__init__(**kwargs)
        vmobjects = [self.language(lang, question_no, **style), *vmobjects]
        self.add(*vmobjects)
        self.question_arrange(aligned_edge=LEFT)

    def language(self, lang, question_no=0, **style):
        text = {"hi": "प्रश्‍न"}.get(lang, "Question")
        if question_no:
            text = "{} {}:".format(text, question_no)
        else:
            text = f"{text} :"

        return Text(text, **style)

    def question_arrange(
        self,
        direction: Vector3D = DOWN,
        buff: float = DEFAULT_MOBJECT_TO_MOBJECT_BUFFER,
        center: bool = True,
        **kwargs,
    ) -> Self:
        self.submobjects[1].next_to(self.submobjects[0], RIGHT, buff, aligned_edge=UP)

        for m1, m2 in zip(self.submobjects[1:], self.submobjects[2:]):
            m2.next_to(m1, direction, buff, **kwargs)
            align = (m1.align if hasattr(m1, "align") else m1.name).lower()
            match align:
                case "cue_column" | "cue-column" | "cue":
                    m1.to_edge(LEFT, buff=5 + buff)
                case "r" | "right":
                    m1.to_edge(RIGHT, buff=-(5 - buff))
                case "c" | "center":
                    m1.set_x(buff + 5)

        if center:
            self.center()
        return self


class AGroup(VMobject):
    def __init__(
        self,
        *vmobjects,
        solve=True,
        lang="en",
        style={"font_size": 25, "color": GREEN_E, "weight": BOLD},
        **kwargs,
    ):
        super().__init__(**kwargs)
        vmobjects = [self.language(lang, solve, **style), *vmobjects]

        self.add(*vmobjects)
        self.question_arrange(aligned_edge=LEFT)

    def language(self, lang, solve=False, **style):
        text = ""
        if solve:
            text = {"hi": "हल"}.get(lang, "Solution")
        else:
            text = {"hi": "उत्तर"}.get(lang, "Answer")

        return Text(f"{text} :", **style)

    def question_arrange(
        self,
        direction: Vector3D = DOWN,
        buff: float = DEFAULT_MOBJECT_TO_MOBJECT_BUFFER,
        center: bool = True,
        **kwargs,
    ) -> Self:
        self.submobjects[1].next_to(self.submobjects[0], RIGHT, buff, aligned_edge=UP)

        for m1, m2 in zip(self.submobjects[1:], self.submobjects[2:]):
            m2.next_to(m1, direction, buff, **kwargs)

            align = (m1.align if hasattr(m1, "align") else m1.name).lower()
            match align:
                case "cue_column" | "cue-column" | "cue":
                    m1.to_edge(LEFT, buff=5 + buff)
                case "r" | "right":
                    m1.to_edge(RIGHT, buff=-(5 - buff))
                case "c" | "center":
                    m1.set_x(buff + 5)

        if center:
            self.center()
        return self
